from ..models import State, Chamber, simple_numbered_districts

NM = State(
    name="New Mexico",
    abbr="NM",
    capital="Santa Fe",
    capital_tz="America/Denver",
    fips="35",
    unicameral=False,
    legislature_name="New Mexico Legislature",
    legislature_organization_id="ocd-organization/7a5a12b7-e955-4999-a469-df3748934953",
    executive_name="Office of the Governor",
    executive_organization_id="ocd-organization/36f78a21-d16b-42a3-9b14-5265b4ebf59e",
    division_id="ocd-division/country:us/state:nm",
    jurisdiction_id="ocd-jurisdiction/country:us/state:nm/government",
    url="https://www.nmlegis.gov",
    lower=Chamber(
        chamber_type="lower",
        name="House",
        organization_id="ocd-organization/af9a59f2-c6b2-4bb1-b270-85f5d40605b4",
        num_seats=70,
        title="Representative",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:nm", "lower", 70
        ),
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/a6719ff4-d604-4144-95e3-453a03672dfd",
        num_seats=42,
        title="Senator",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:nm", "upper", 42
        ),
    ),
)



from ast import literal_eval
import datetime
import jwt
import logging
import logging.config
import os
import requests
from requests.adapters import HTTPAdapter
from urllib3.util import Retry
import time
from typing import List, Dict

from .. import settings


class MetricTypes:
    CounterType = "count"
    GaugeType = "gauge"
    TimingType = "timing"
    SetType = "set"


class Instrumentation(object):
    def __init__(self) -> None:
        """
        We're currently leveraging https://github.com/civic-eagle/statsd-http-proxy
        to have an easy to reason about stats aggregation tool
        on the other end of an internet connection

        Essentially statsd-style metrics, we can rely on the aggregator
        to continually emit these stats so we get useful stats for monitoring/reporting
        """
        logging.config.dictConfig(settings.LOGGING)
        self.logger = logging.getLogger("openstates.stats")
        # use a literal_eval to properly turn a string into a bool (literal_eval 'cause it's safer than stdlib eval)
        self.enabled = literal_eval(os.environ.get("STATS_ENABLED", "False"))
        if not self.enabled:
            self.logger.debug("Stat emission is not enabled.")
            return
        token: str = self._jwt_token()
        self._batch: List[Dict] = list()
        self.prefix: str = os.environ.get("STATS_PREFIX", "openstates_")
        self.endpoint: str = os.environ.get("STATS_ENDPOINT", "")
        if self.endpoint.endswith("/"):
            self.endpoint = self.endpoint.strip("/")
        stats_retries: int = int(os.environ.get("STATS_RETRIES", 3))
        self.batch_size: int = int(os.environ.get("STATS_BATCH_SIZE", 50))
        self.default_tags: Dict = dict()
        headers = {"Content-Type": "application/json"}
        if token:
            headers["X-JWT-Token"] = token
        retry = Retry(
            total=stats_retries,
            read=stats_retries,
            connect=stats_retries,
            backoff_factor=0.3,
            status_forcelist=(500, 502, 503, 504, 429, 104),
        )
        adapter = HTTPAdapter(max_retries=retry)
        self._stat_client = requests.Session()
        self._stat_client.headers.update(headers)
        # only mount https endpoint if we need it
        if self.endpoint.startswith("https"):
            self._stat_client.mount("https://", adapter)
        else:
            self._stat_client.mount("http://", adapter)
        self.logger.debug(
            f"Stats emission to {self.endpoint} configured with batch size: {self.batch_size}"
        )

    def _jwt_token(self) -> str:
        """
        Generate a JWT token from a consistent environment secret
        Return empty strings when generation isn't possible.
        """
        if not self.enabled:
            return ""
        secret = os.environ.get("STATS_JWT_SECRET", "")
        if not secret:
            return ""
        """
        We need a pretty long expiration time for runs that take 12+ hours.
        So we add the `exp` setting as 36 hours out. Should give us plenty of time.
        """
        return jwt.encode(
            {
                "id": "openstates",
                "exp": datetime.datetime.now(tz=datetime.timezone.utc)
                + datetime.timedelta(hours=36),
            },
            secret,
            algorithm="HS256",
        )

    def _send_stats(self, force: bool = False) -> None:
        """
        Needs to be broken out from _process_metric to have a
        "write at shutdown" function (force=True)
        Even with force=True, we should only write when there is data
        to write. Otherwise, just skip things.
        """
        batch_len = len(self._batch)
        if (force and batch_len > 0) or batch_len > self.batch_size:
            if not self.enabled:
                self.logger.debug("Stats diasbled. Skipping send")
                return
            if not self.endpoint:
                self.logger.debug("No stats endpoint defined. Not emitting stats")
                return
            self.logger.debug(f"Sending stats batch: {self._batch}")
            try:
                self._stat_client.post(f"{self.endpoint}/batch", json=self._batch)
                self._batch = list()
            except Exception as e:
                self.logger.warning(
                    f"Failed to write {len(self._batch)} stats to {self.endpoint} :: {e}"
                )

    def _process_metric(
        self,
        metric_type: str,
        metric: str,
        tags: dict,
        value: float,
        sample_rate: float = 0,
    ) -> None:
        """
        Ensure consistent formatting of data objects to add to batch for sending
        """
        if not self.enabled:
            return
        # apply defaults only if not overridden
        for k, v in self.default_tags.items():
            if k not in tags:
                tags[k] = v

        """
        Statsd http proxy specific formatting
        """
        # list(set()) to remove duplicates
        tagstr = ",".join(list(set([f"{k}={v}" for k, v in tags.items()])))
        data = {
            "value": value,
            "metric": f"{self.prefix}{metric}",
            "metric_type": metric_type,
            "tags": tagstr,
        }
        if sample_rate:
            data["sampleRate"] = sample_rate
        # End specific formatting

        self._batch.append(data)
        """
        we attempt to send (without forcing) after
        adding each stat to make sure we emit
        batches as quickly as we can, without
        overloading the write endpoint with tiny writes
        """
        self._send_stats()

    """
    Wrapper scripts for easier sending

    Using these functions would look like:

    from utils.instrument import Instrumentation
    stats = Instrumentation()
    stats.send_gauge("objects_scraped", 10, [{"jurisdiction": "ca"}])
    stats.close()
    """

    def close(self) -> None:
        """
        "Shut down" our instrumentation connection
        Currently just forces any batched stats out
        Keep as a wrapper script for later extensibility
        """
        if not self.enabled:
            return
        self._send_stats(force=True)

    def send_last_run(self, metric: str, tags: dict = {}) -> None:
        """
        Set a gauge with a current timestamp
        Emulates a "last run time" feature simply
        """
        if not self.enabled:
            return
        self._process_metric(MetricTypes.GaugeType, metric, tags, int(time.time()))

    def send_counter(
        self,
        metric: str,
        value: float,
        tags: dict = {},
        sample_rate: float = 0,
    ) -> None:
        if not self.enabled:
            return
        self._process_metric(MetricTypes.CounterType, metric, tags, value, sample_rate)

    def send_gauge(self, metric: str, value: float, tags: dict = {}) -> None:
        """
        Gauges aren't calculated, so don't require a sample rate
        """
        if not self.enabled:
            return
        self._process_metric(MetricTypes.GaugeType, metric, tags, value)

    def send_timing(
        self,
        metric: str,
        value: float,
        tags: dict = {},
        sample_rate: float = 0,
    ) -> None:
        if not self.enabled:
            return
        self._process_metric(MetricTypes.TimingType, metric, tags, value)

    def send_set(self, metric: str, value: float, tags: dict = {}) -> None:
        """
        Sets (naturally) aren't calculated, so don't require a sample rate
        """
        if not self.enabled:
            return
        self._process_metric(MetricTypes.SetType, metric, tags, value)



from . import states
from .dc import DC
from .pr import PR
from .us import US

STATES = states.STATES
STATES_AND_DC_PR = STATES + [DC, PR]

STATES_BY_ABBR = {s.abbr: s for s in STATES_AND_DC_PR + [US]}
STATES_BY_JID = {s.jurisdiction_id: s for s in STATES_AND_DC_PR + [US]}
STATES_BY_NAME = {s.name.lower(): s for s in STATES_AND_DC_PR + [US]}



import re
import tempfile
import functools
import subprocess
from lxml import html  # type: ignore


def pdfdata_to_text(data: bytes) -> str:
    with tempfile.NamedTemporaryFile(delete=True) as tmpf:
        tmpf.write(data)
        tmpf.flush()
        try:
            pipe = subprocess.Popen(
                ["pdftotext", "-layout", tmpf.name, "-"],
                stdout=subprocess.PIPE,
                close_fds=True,
            ).stdout
        except OSError as e:
            raise EnvironmentError(
                f"error running pdftotext, missing executable? [{e}]"
            )
        if not pipe:
            raise EnvironmentError("could not open pipe")
        data = pipe.read()
        pipe.close()
        return data.decode("utf8", "ignore")


def clean(text: str) -> str:
    text = text.replace("\xa0", " ")  # nbsp -> sp
    text = text.replace("\r\n", "\n")  # replace carriage returns
    text = re.sub(r"[ \t]", " ", text)  # collapse spaces
    # collapse newlines too?
    return text


def _text_near_line_numbers(lines: str, regex: str) -> str:
    """ used for before & after line numbers """
    text = []
    for line in lines.splitlines():
        # real bill text starts with an optional space, line number,
        # more spaces, then real text
        match = re.match(regex, line)
        if match:
            text.append(match.group(1))

    # return all real bill text joined w/ newlines
    return "\n".join(text)


text_after_line_numbers = functools.partial(
    _text_near_line_numbers, regex=r"\s*\d+\s+(.*)"
)
text_before_line_numbers = functools.partial(
    _text_near_line_numbers, regex=r"(.*?)\s+\d+\s*"
)


def text_from_element_lxml(data: bytes, lxml_query: str) -> str:
    html_document = html.fromstring(data)
    matching_elements = html_document.findall(lxml_query)

    # To ensure that we exit non-zero if there are multiple matching elements
    # on the page, raise an exception: this means that the extraction
    # code needs to be updated.
    assert (
        len(matching_elements) == 1
    ), f"{len(matching_elements)} matches for {lxml_query}"

    text_inside_element = matching_elements[0].text_content()
    return text_inside_element


def text_from_element_xpath(data: bytes, lxml_xpath_query: str) -> str:
    html_document = html.fromstring(data)
    matching_elements = html_document.xpath(lxml_xpath_query)

    # To ensure that we exit non-zero if there are multiple matching elements
    # on the page, raise an exception: this means that the extraction
    # code needs to be updated.
    assert (
        len(matching_elements) == 1
    ), f"{len(matching_elements)} matches for {lxml_xpath_query}"

    text_inside_element = matching_elements[0].text_content()
    return text_inside_element


def text_from_element_siblings_lxml(data: bytes, lxml_query: str) -> str:
    html_document = html.fromstring(data)
    matching_elements = html_document.findall(lxml_query)

    text_inside_elements = ""
    for element in matching_elements:
        text_inside_elements += element.text_content() + "\n"

    return text_inside_elements


def text_from_element_siblings_xpath(data: bytes, lxml_query: str) -> str:
    html_document = html.fromstring(data)
    matching_elements = html_document.xpath(lxml_query)

    text_inside_elements = ""
    for element in matching_elements:
        text_inside_elements += element.text_content() + "\n"

    return text_inside_elements



# Generated by Django 3.2.2 on 2021-10-26 17:38

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("data", "0042_delete_personcontactdetail"),
    ]

    operations = [
        migrations.AddField(
            model_name="legislativesession",
            name="active",
            field=models.BooleanField(default=False),
        ),
    ]



from .common import extractor_for_elements_by_xpath, extract_line_numbered_pdf, Metadata

# the sample.csv includes only MSWord files for some Delaware laws
# (actually "HCR" resolutions honoring folks)
# in 'real life' we'll want to make sure to feed to this scraper the PDF link,
# rather than the Docx link.
# Docxes are ignored, PDFs will be handled IFF 'HCR' is in the title.


def handle_delaware(data: bytes, metadata: Metadata) -> str:
    if metadata["media_type"] == "text/html" and "HCR" not in metadata["title"]:
        return extractor_for_elements_by_xpath("/html/body/div[2] | /html/body/div[3]")(
            data, metadata
        )
    elif metadata["media_type"] == "application/pdf" and "HCR" not in metadata["title"]:
        # Del., like many states, appears to publish all bills as both text and HTML
        # so we don't *need* to extract from PDF.
        return ""
    elif metadata["media_type"] == "application/pdf" and "HCR" in metadata["title"]:
        # haha, but Delaware does NOT appear to publish concurrent resolutions
        # (honoring people and recognizing the contributions of general aviation, etc.)
        # as HTML. Just PDF and MSWord.
        # so we handle these here and ignore docx (below)
        return extract_line_numbered_pdf(data, metadata)
    else:
        print("unexpected media type in delaware", metadata["media_type"])
        assert False



import re
import uuid
from django.db import models
from django.core.validators import RegexValidator
from .. import common


class OCDIDField(models.CharField):
    def __init__(self, *args, **kwargs):
        self.ocd_type = kwargs.pop("ocd_type")
        if self.ocd_type != "jurisdiction":
            kwargs["default"] = lambda: "ocd-{}/{}".format(self.ocd_type, uuid.uuid4())
            # len('ocd-') + len(ocd_type) + len('/') + len(uuid)
            #       = 4 + len(ocd_type) + 1 + 36
            #       = len(ocd_type) + 41
            kwargs["max_length"] = 41 + len(self.ocd_type)
            regex = (
                "^ocd-" + self.ocd_type + "/[0-9a-f]{8}-([0-9a-f]{4}-){3}[0-9a-f]{12}$"
            )
        else:
            kwargs["max_length"] = 300
            regex = common.JURISDICTION_ID_REGEX

        kwargs["primary_key"] = True
        # get pattern property if it exists, otherwise just return the object (hopefully a string)
        msg = "ID must match " + getattr(regex, "pattern", regex)
        kwargs["validators"] = [RegexValidator(regex=regex, message=msg, flags=re.U)]
        super(OCDIDField, self).__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super(OCDIDField, self).deconstruct()
        if self.ocd_type != "jurisdiction":
            kwargs.pop("default")
        kwargs.pop("max_length")
        kwargs.pop("primary_key")
        kwargs["ocd_type"] = self.ocd_type
        return (name, path, args, kwargs)


class OCDBase(models.Model):
    """common base fields across all top-level models"""

    created_at = models.DateTimeField(
        auto_now_add=True, help_text="The date and time of creation."
    )
    updated_at = models.DateTimeField(
        auto_now=True, help_text="The date and time of the last update."
    )
    extras = models.JSONField(
        default=dict,
        blank=True,
        help_text="A key-value store for storing arbitrary information not covered elsewhere.",
    )

    class Meta:
        abstract = True


class RelatedBase(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)

    class Meta:
        abstract = True


class LinkBase(RelatedBase):
    note = models.CharField(
        max_length=300,
        blank=True,
        help_text="A short, optional note related to an object.",
    )
    url = models.URLField(
        max_length=2000, help_text="A hyperlink related to an object."
    )

    class Meta:
        abstract = True

    def __str__(self):
        return self.url


class MimetypeLinkBase(RelatedBase):
    media_type = models.CharField(max_length=100)
    url = models.URLField(max_length=2000)

    class Meta:
        abstract = True


class IdentifierBase(RelatedBase):
    identifier = models.CharField(
        max_length=300,
        help_text="A unique identifier developed by an upstream or third party source.",
    )
    scheme = models.CharField(
        max_length=300, help_text="The name of the service that created the identifier."
    )

    class Meta:
        abstract = True

    def __str__(self):
        return self.identifier


class RelatedEntityBase(RelatedBase):
    name = models.CharField(max_length=2000)
    entity_type = models.CharField(max_length=20, blank=True)

    # optionally tied to an organization or person if it was linkable
    # for these two on_delete is SET_NULL so that deletion of a linked entity doesn't
    # delete this object- it should instead just become unresolved (NULL)
    organization = models.ForeignKey(
        "data.Organization", null=True, on_delete=models.SET_NULL
    )
    person = models.ForeignKey("data.Person", null=True, on_delete=models.SET_NULL)

    @property
    def entity_name(self):
        if self.entity_type == "organization" and self.organization_id:
            return self.organization.name
        elif self.entity_type == "person" and self.person_id:
            return self.person.name
        else:
            return self.name

    @property
    def entity_id(self):
        if self.entity_type == "organization":
            return self.organization_id
        if self.entity_type == "person":
            return self.person_id
        return None

    class Meta:
        abstract = True



# Generated by Django 3.0.5 on 2020-08-11 10:09

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("data", "0014_remove_person_current_role_division_id"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="billaction",
            name="extras",
        ),
        migrations.AlterIndexTogether(
            name="membership",
            index_together={("organization", "person", "post")},
        ),
        migrations.RemoveField(
            model_name="membership",
            name="label",
        ),
    ]



from ..models import State, Chamber, simple_numbered_districts

CA = State(
    name="California",
    abbr="CA",
    capital="Sacramento",
    capital_tz="America/Los_Angeles",
    fips="06",
    unicameral=False,
    legislature_name="California State Legislature",
    legislature_organization_id="ocd-organization/f1c6eb47-c439-42e5-9b6f-d7f6121f20e9",
    executive_name="Office of the Governor",
    executive_organization_id="ocd-organization/b7f8d62d-b8f0-4eba-8de3-a42c1f64f9fe",
    division_id="ocd-division/country:us/state:ca",
    jurisdiction_id="ocd-jurisdiction/country:us/state:ca/government",
    url="http://www.legislature.ca.gov/",
    lower=Chamber(
        chamber_type="lower",
        name="Assembly",
        organization_id="ocd-organization/e70e812e-44d2-46b0-9eae-2f3b9a1d5150",
        num_seats=80,
        title="Assemblymember",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:ca", "lower", 80
        ),
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/98623fa0-f9c3-47ae-80bb-51465349ed71",
        num_seats=40,
        title="Senator",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:ca", "upper", 40
        ),
    ),
)



from ..models import State, Chamber, simple_numbered_districts

WV = State(
    name="West Virginia",
    abbr="WV",
    capital="Charleston",
    capital_tz="America/New_York",
    fips="54",
    unicameral=False,
    legislature_name="West Virginia Legislature",
    legislature_organization_id="ocd-organization/237dc1a5-2314-433e-887e-d63a970cf67d",
    executive_name="Office of the Governor",
    executive_organization_id="ocd-organization/193c7d7d-3161-5a34-a7b1-df72b553b7f3",
    division_id="ocd-division/country:us/state:wv",
    jurisdiction_id="ocd-jurisdiction/country:us/state:wv/government",
    url="http://www.legis.state.wv.us/",
    lower=Chamber(
        chamber_type="lower",
        name="House",
        organization_id="ocd-organization/3fcd7789-fa25-4d8b-a4b9-7c0d1d0f8bc4",
        num_seats=100,
        title="Delegate",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:wv", "lower", 100
        ),
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/660bc141-f4cb-4d98-b334-b12b637776ba",
        num_seats=34,
        title="Senator",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:wv",
            "upper",
            17,
            num_seats=2,
        ),
    ),
)



# Generated by Django 3.2.2 on 2021-10-13 19:07

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("data", "0041_personoffice"),
    ]

    operations = [
        migrations.DeleteModel(
            name="PersonContactDetail",
        ),
    ]



# Generated by Django 3.0.5 on 2020-09-14 10:32

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("data", "0017_remove_person_summary"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="bill",
            name="locked_fields",
        ),
        migrations.RemoveField(
            model_name="event",
            name="locked_fields",
        ),
        migrations.RemoveField(
            model_name="jurisdiction",
            name="locked_fields",
        ),
        migrations.RemoveField(
            model_name="membership",
            name="locked_fields",
        ),
        migrations.RemoveField(
            model_name="organization",
            name="locked_fields",
        ),
        migrations.RemoveField(
            model_name="person",
            name="locked_fields",
        ),
        migrations.RemoveField(
            model_name="post",
            name="locked_fields",
        ),
        migrations.RemoveField(
            model_name="voteevent",
            name="locked_fields",
        ),
    ]



import pytest  # type: ignore
from openstates.utils.transformers import fix_bill_id, collapse_whitespace


@pytest.mark.parametrize(
    "orig,exp",
    [
        ("HB1", "HB 1"),
        ("   HB1  ", "HB 1"),
        ("HB 0001", "HB 1"),
        ("hb    1", "HB 1"),
        ("SJRA", "SJR A"),
    ],
)
def test_fix_bill_id(orig: str, exp: str) -> None:
    assert fix_bill_id(orig) == exp


@pytest.mark.parametrize(
    "orig,exp",
    [
        ("HR 1", "HR 1"),
        ("HJRES 1", "HJRES 1"),
        ("HCONRES 1", "HCONRES 1"),
        ("HRES 1", "HRES 1"),
        ("S 1", "S 1"),
        ("SJRES 1", "SJRES 1"),
        ("SCONRES 1", "SCONRES 1"),
        ("SRES 1", "SRES 1"),
    ],
)
def test_fix_bill_id_federal(orig: str, exp: str) -> None:
    assert fix_bill_id(orig) == exp


@pytest.mark.parametrize(
    "orig,exp",
    [
        ("Simple", "Simple"),
        ("One Space", "One Space"),
        ("Two  Spaces", "Two Spaces"),
        ("Line\nBreak", "Line Break"),
        ("Many  Breaks \rto  \t  Fix", "Many Breaks to Fix"),
    ],
)
def test_collapse_whitespace(orig: str, exp: str) -> None:
    assert collapse_whitespace(orig) == exp



import copy
from .base import BaseModel, SourceMixin, LinkMixin
from .schemas.organization import schema as org_schema
from ..utils import _make_pseudo_id
from ..exceptions import ScrapeValueError

# a copy of the org schema without sources
org_schema_no_sources = copy.deepcopy(org_schema)
org_schema_no_sources["properties"].pop("sources")


class Organization(BaseModel, SourceMixin, LinkMixin):
    """
    A single popolo-style Organization
    """

    _type = "organization"
    _schema = org_schema

    def __init__(self, name, *, classification="", parent_id=None, chamber=None):
        """
        Constructor for the Organization object.
        """
        super(Organization, self).__init__()
        self.name = name
        self.classification = classification
        self.parent_id = pseudo_organization(parent_id, chamber)

    def __str__(self):
        return self.name

    def validate(self):
        schema = None
        # these are implicitly declared & do not require sources
        if self.classification in (
            "party",
            "legislature",
            "upper",
            "lower",
            "executive",
        ):
            schema = org_schema_no_sources
        return super(Organization, self).validate(schema=schema)

    def add_post(self, label, role, **kwargs):
        # STUB: will be removed soon
        pass

    def add_member(self, name_or_person, role="member", **kwargs):
        # STUB: will be removed soon
        pass


def pseudo_organization(organization, classification, default=None):
    """helper for setting an appropriate ID for organizations"""
    if organization and classification:
        raise ScrapeValueError("cannot specify both classification and organization")
    elif classification:
        return _make_pseudo_id(classification=classification)
    elif organization:
        if isinstance(organization, Organization):
            return organization._id
        elif isinstance(organization, str):
            return organization
        else:
            return _make_pseudo_id(**organization)
    elif default is not None:
        return _make_pseudo_id(classification=default)
    else:
        return None



import re
import os
import uuid
import typing
import yaml
from pathlib import Path
from enum import Enum
from ... import metadata
from ...models.common import BaseModel


class EnumDumper(yaml.SafeDumper):
    def represent_data(self, data):  # type: ignore
        if isinstance(data, Enum):
            return self.represent_data(data.value)
        return super().represent_data(data)


def ocd_uuid(type: str) -> str:
    return "ocd-{}/{}".format(type, uuid.uuid4())


def get_base_path() -> Path:
    # there are two options for where the people utilities look for their data:

    if "OS_PEOPLE_DIRECTORY" in os.environ:
        # 1. use the environment variable OS_PEOPLE_DIRECTORY
        base_dir = Path(os.environ["OS_PEOPLE_DIRECTORY"])
    else:
        # 2. if not set, we'll look in a sibling directory
        # this is a nice fallback for local development, but not intended to work
        # in production environments
        sibling_directory = Path(__file__).parents[4] / "people"
        if sibling_directory.exists():
            base_dir = sibling_directory
        else:
            raise EnvironmentError(
                "could not find openstates/people checkout, set OS_PEOPLE_DIRECTORY env variable"
            )

    if not (base_dir / "data").exists():
        raise EnvironmentError(
            f"{base_dir}/data does not exist, check OS_PEOPLE_DIRECTORY"
        )

    return base_dir


def get_data_path(abbr: str) -> Path:
    return get_base_path() / "data" / abbr


def get_all_abbreviations() -> list[str]:
    return sorted(x.name for x in (get_base_path() / "data").iterdir())


def dump_obj(
    obj: BaseModel,
    *,
    output_dir: typing.Optional[Path] = None,
    filename: typing.Union[Path, str, None] = None,
) -> None:
    obj_dict = obj.to_dict()
    if output_dir:
        filename = output_dir / get_new_filename(obj_dict)
    if not filename:
        raise ValueError("must provide output_dir or filename parameter")
    with open(filename, "w") as f:
        yaml.dump(
            obj_dict,
            f,
            default_flow_style=False,
            sort_keys=False,
            Dumper=EnumDumper,
        )


def get_new_filename(obj: dict) -> str:
    id = obj["id"].split("/")[1]
    name = obj["name"]
    name = re.sub(r"\s+", "-", name)
    name = re.sub(r"[^a-zA-Z-]", "", name)
    return f"{name}-{id}.yml"


def legacy_districts(
    abbr: typing.Optional[str] = None, jurisdiction_id: typing.Optional[str] = None
) -> dict[str, list[str]]:
    """ can take jurisdiction_id or abbr via kwargs """
    legacy_districts: dict[str, list[str]] = {"upper": [], "lower": []}
    for d in metadata.lookup(
        abbr=abbr, jurisdiction_id=jurisdiction_id
    ).legacy_districts:
        legacy_districts[d.chamber_type].append(d.name)
    return legacy_districts


def load_municipalities(abbr: str) -> list[dict]:
    try:
        with open(get_data_path(abbr) / "municipalities.yml") as f:
            return typing.cast(list, yaml.safe_load(f))
    except FileNotFoundError:
        return []



# Generated by Django 3.2 on 2021-04-21 15:38

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("data", "0032_event_dedupe_key"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="organizationsource",
            name="organization",
        ),
        migrations.AddField(
            model_name="organization",
            name="links",
            field=django.contrib.postgres.fields.jsonb.JSONField(
                default=list, blank=True
            ),
        ),
        migrations.AddField(
            model_name="organization",
            name="sources",
            field=django.contrib.postgres.fields.jsonb.JSONField(
                default=list, blank=True
            ),
        ),
        migrations.DeleteModel(
            name="OrganizationLink",
        ),
        migrations.DeleteModel(
            name="OrganizationSource",
        ),
    ]



import pytest  # type: ignore
import datetime
from pathlib import Path
from openstates.utils.people.lint_people import (
    validate_name,
    validate_roles,
    validate_offices,
    get_expected_districts,
    compare_districts,
    Validator,
    BadVacancy,
    PersonType,
)  # noqa
from openstates.models.people import Person, Role, Party, Office


EXAMPLE_OCD_JURISDICTION_ID = "ocd-jurisdiction/country:us/government"
EXAMPLE_OCD_PERSON_ID = "ocd-person/12345678-0000-1111-2222-1234567890ab"
EXAMPLE_OCD_ORG_ID = "ocd-organization/00001111-2222-3333-aaaa-444455556666"


@pytest.mark.parametrize(
    "person,expected",
    [
        (Person(id=EXAMPLE_OCD_PERSON_ID, name="Phillip J Swoozle", roles=[]), []),
        (
            Person(id=EXAMPLE_OCD_PERSON_ID, name="Phillip Swoozle", roles=[]),
            [
                "missing given_name that could be set to 'Phillip', run with --fix",
                "missing family_name that could be set to 'Swoozle', run with --fix",
            ],
        ),
        (
            Person(
                id=EXAMPLE_OCD_PERSON_ID,
                name="Phillip Swoozle",
                given_name="Phil",
                roles=[],
            ),
            [
                "missing family_name that could be set to 'Swoozle', run with --fix",
            ],
        ),
        (
            Person(
                id=EXAMPLE_OCD_PERSON_ID,
                name="Phillip Swoozle",
                given_name="Phil",
                family_name="Swoozle",
                roles=[],
            ),
            [],
        ),
    ],
)
def test_validate_name_errors(person, expected):
    assert validate_name(person, PersonType.LEGISLATIVE, fix=False).errors == expected
    assert validate_name(person, PersonType.LEGISLATIVE, fix=False).warnings == []
    assert validate_name(person, PersonType.LEGISLATIVE, fix=False).fixes == []


def test_validate_name_fixes():
    person = Person(id=EXAMPLE_OCD_PERSON_ID, name="Phillip Swoozle", roles=[])
    result = validate_name(person, PersonType.LEGISLATIVE, fix=True)
    assert result.errors == []
    assert len(result.fixes) == 2
    assert person.given_name == "Phillip"
    assert person.family_name == "Swoozle"

    # no fixes on an OK name
    result = validate_name(person, PersonType.LEGISLATIVE, fix=True)
    assert result.errors == result.fixes == []


@pytest.mark.parametrize(
    "roles,expected",
    [
        (
            [Role(type="lower", jurisdiction=EXAMPLE_OCD_JURISDICTION_ID, district=3)],
            [],
        ),
        (
            [
                Role(
                    type="upper", jurisdiction=EXAMPLE_OCD_JURISDICTION_ID, district=3
                ),
                Role(
                    type="lower", jurisdiction=EXAMPLE_OCD_JURISDICTION_ID, district=3
                ),
            ],
            ["2 active roles"],
        ),
        ([], ["no active roles"]),
        (
            [
                Role(
                    type="governor",
                    jurisdiction=EXAMPLE_OCD_JURISDICTION_ID,
                    end_date="1990",
                )
            ],
            ["no active roles"],
        ),
    ],
)
def test_validate_roles_roles(roles, expected):
    person = Person(
        id=EXAMPLE_OCD_PERSON_ID,
        name="Example Person",
        roles=roles,
        party=[Party(name="Republican")],
    )
    assert validate_roles(person, "roles") == expected


@pytest.mark.parametrize(
    "roles,expected",
    [
        (
            [Role(type="lower", jurisdiction=EXAMPLE_OCD_JURISDICTION_ID, district=3)],
            ["1 active roles on retired person"],
        ),
        (
            [
                Role(
                    type="upper", jurisdiction=EXAMPLE_OCD_JURISDICTION_ID, district=3
                ),
                Role(
                    type="lower", jurisdiction=EXAMPLE_OCD_JURISDICTION_ID, district=3
                ),
            ],
            ["2 active roles on retired person"],
        ),
        ([], []),
        (
            [
                Role(
                    type="governor",
                    jurisdiction=EXAMPLE_OCD_JURISDICTION_ID,
                    end_date="1990",
                )
            ],
            [],
        ),
    ],
)
def test_validate_roles_retired(roles, expected):
    person = Person(
        id=EXAMPLE_OCD_PERSON_ID,
        name="Example Person",
        roles=roles,
        party=[Party(name="Republican")],
    )
    assert validate_roles(person, "roles", retired=True) == expected


@pytest.mark.parametrize(
    "offices,expected",
    [
        ([], []),
        (
            [
                {"classification": "capitol", "voice": "111-222-3333"},
                {"classification": "capitol", "fax": "111-222-5555"},
            ],
            ["Multiple capitol offices, condense to one."],
        ),
        (
            [
                {"classification": "district", "voice": "111-222-4333"},
                {"classification": "district", "voice": "555-555-5555"},
            ],
            [],
        ),
        (
            [
                {"classification": "district", "voice": "123-444-4444"},
                {"classification": "capitol", "voice": "123-444-4444"},
            ],
            [
                "Value '123-444-4444' used multiple times: district voice and capitol voice"
            ],
        ),
    ],
)
def test_validate_offices(offices, expected):
    person = Person(
        id=EXAMPLE_OCD_PERSON_ID,
        name="Example Person",
        roles=[],
        offices=[Office(**c) for c in offices],
    )
    assert validate_offices(person) == expected


def test_get_expected_districts():
    expected = get_expected_districts({}, "ne")
    assert len(expected["legislature"]) == 49
    assert expected["legislature"]["1"] == 1

    expected = get_expected_districts({}, "md")
    print(expected)
    assert expected["lower"]["43A"] == 2
    assert expected["lower"]["43B"] == 1


def test_expected_districts_vacancies():
    vacancies = {
        "ne": {
            "vacancies": [
                {
                    "chamber": "legislature",
                    "district": "1",
                    "vacant_until": datetime.date(2100, 1, 1),
                }
            ]
        }
    }
    expected = get_expected_districts(vacancies, "ne")
    assert expected["legislature"]["1"] == 0

    with pytest.raises(BadVacancy):
        get_expected_districts(
            {
                "ne": {
                    "vacancies": [
                        {
                            "chamber": "upper",
                            "district": "2",
                            "vacant_until": datetime.date(2000, 1, 1),
                        }
                    ]
                }
            },
            "ne",
        )


@pytest.mark.parametrize(
    "expected,actual,errors",
    [
        ({"A": 1, "B": 1}, {"A": ["a"], "B": ["a"]}, 0),  # good
        ({"A": 1}, {"A": ["a"], "B": ["a"]}, 1),  # extra district
        ({"A": 1, "B": 1}, {"A": ["a"]}, 1),  # missing district
        ({"A": 1, "B": 1}, {"A": [], "B": ["a"]}, 1),  # missing leg
    ],
)
def test_compare_districts(expected, actual, errors):
    e = compare_districts({"upper": expected}, {"upper": actual})
    assert len(e) == errors


def test_compare_districts_overfill():
    expected = {"A": 1}
    actual = {"A": ["Anne", "Bob"]}
    e = compare_districts({"upper": expected}, {"upper": actual})
    assert len(e) == 1
    assert "Anne" in e[0]
    assert "Bob" in e[0]


def test_person_duplicates():
    v = Validator("ak", {}, False, False)

    people = [
        # duplicates across people
        {
            "id": "ocd-person/11111111-1111-1111-1111-111111111111",
            "name": "One",
            "ids": {"twitter": "no-twitter", "youtube": "fake"},
            "roles": [],
        },
        {
            "id": "ocd-person/22222222-2222-2222-2222-222222222222",
            "name": "Two",
            "ids": {"twitter": "no-twitter", "youtube": "fake"},
            "roles": [],
        },
        # duplicate on same person
        {
            "id": "ocd-person/33333333-3333-3333-3333-333333333333",
            "name": "Three",
            "ids": {"twitter": "no-twitter"},
            "other_identifiers": [
                {"scheme": "external_service_id", "identifier": "XYZ"},
                {"scheme": "external_service_id", "identifier": "XYZ"},
            ],
            "roles": [],
        },
        {
            "id": "ocd-person/44444444-4444-4444-4444-444444444444",
            "name": "Four",
            "ids": {"twitter": "no-twitter"},
            "roles": [],
        },
    ]
    for p in people:
        v.validate_person(p, Path(p["name"] + ".yml"), PersonType.LEGISLATIVE)
    errors = v.check_duplicates()
    assert len(errors) == 3
    assert 'duplicate youtube: "fake" One.yml, Two.yml' in errors
    assert 'duplicate external_service_id: "XYZ" Three.yml, Three.yml' in errors
    assert (
        'duplicate twitter: "no-twitter" One.yml, Two.yml, Three.yml and 1 more...'
        in errors
    )


def test_filename_id_test():
    person = dict(
        id=EXAMPLE_OCD_PERSON_ID,
        name="Jane Smith",
        roles=[],
    )
    v = Validator("ak", {}, False, False)
    v.validate_person(person, Path("bad-filename"), PersonType.LEGISLATIVE)
    for err in v.errors["bad-filename"]:
        if "not in filename" in err:
            break
    else:
        raise AssertionError("did not check for id in filename")



from ..models import State, Chamber, simple_numbered_districts

FL = State(
    name="Florida",
    abbr="FL",
    capital="Tallahassee",
    capital_tz="America/New_York",
    fips="12",
    unicameral=False,
    legislature_name="Florida Legislature",
    legislature_organization_id="ocd-organization/a0b3ad5f-c852-483b-b746-444d3d795701",
    executive_name="Office of the Governor",
    executive_organization_id="ocd-organization/392f8763-ef77-5ce1-9b8e-643b129c3bd3",
    division_id="ocd-division/country:us/state:fl",
    jurisdiction_id="ocd-jurisdiction/country:us/state:fl/government",
    url="http://myflorida.com",
    lower=Chamber(
        chamber_type="lower",
        name="House",
        organization_id="ocd-organization/8b0147dc-8942-4fd5-91c4-a72a92a8e1fe",
        num_seats=120,
        title="Representative",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:fl", "lower", 120
        ),
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/c540c1bb-d614-4174-86db-a146526169f8",
        num_seats=40,
        title="Senator",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:fl", "upper", 40
        ),
    ),
)



from ..models import State, Chamber, District

MA = State(
    name="Massachusetts",
    abbr="MA",
    capital="Boston",
    capital_tz="America/New_York",
    fips="25",
    unicameral=False,
    legislature_name="Massachusetts General Court",
    legislature_organization_id="ocd-organization/9c4e7b53-9912-4394-a40a-14b5ecdb8aba",
    executive_name="Office of the Governor",
    executive_organization_id="ocd-organization/61a1385f-f512-4d4b-826f-3118c9a775bf",
    division_id="ocd-division/country:us/state:ma",
    jurisdiction_id="ocd-jurisdiction/country:us/state:ma/government",
    url="http://mass.gov",
    lower=Chamber(
        chamber_type="lower",
        name="House",
        organization_id="ocd-organization/ca38ad9c-c3d5-4c4f-bc2f-d885218ed802",
        num_seats=162,
        title="Representative",
        districts=[
            District(
                "10th Bristol",
                "lower",
                "ocd-division/country:us/state:ma/sldl:10th_bristol",
                1,
            ),
            District(
                "10th Essex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:10th_essex",
                1,
            ),
            District(
                "10th Hampden",
                "lower",
                "ocd-division/country:us/state:ma/sldl:10th_hampden",
                1,
            ),
            District(
                "10th Middlesex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:10th_middlesex",
                1,
            ),
            District(
                "10th Norfolk",
                "lower",
                "ocd-division/country:us/state:ma/sldl:10th_norfolk",
                1,
            ),
            District(
                "10th Plymouth",
                "lower",
                "ocd-division/country:us/state:ma/sldl:10th_plymouth",
                1,
            ),
            District(
                "10th Suffolk",
                "lower",
                "ocd-division/country:us/state:ma/sldl:10th_suffolk",
                1,
            ),
            District(
                "10th Worcester",
                "lower",
                "ocd-division/country:us/state:ma/sldl:10th_worcester",
                1,
            ),
            District(
                "11th Bristol",
                "lower",
                "ocd-division/country:us/state:ma/sldl:11th_bristol",
                1,
            ),
            District(
                "11th Essex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:11th_essex",
                1,
            ),
            District(
                "11th Hampden",
                "lower",
                "ocd-division/country:us/state:ma/sldl:11th_hampden",
                1,
            ),
            District(
                "11th Middlesex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:11th_middlesex",
                1,
            ),
            District(
                "11th Norfolk",
                "lower",
                "ocd-division/country:us/state:ma/sldl:11th_norfolk",
                1,
            ),
            District(
                "11th Plymouth",
                "lower",
                "ocd-division/country:us/state:ma/sldl:11th_plymouth",
                1,
            ),
            District(
                "11th Suffolk",
                "lower",
                "ocd-division/country:us/state:ma/sldl:11th_suffolk",
                1,
            ),
            District(
                "11th Worcester",
                "lower",
                "ocd-division/country:us/state:ma/sldl:11th_worcester",
                1,
            ),
            District(
                "12th Bristol",
                "lower",
                "ocd-division/country:us/state:ma/sldl:12th_bristol",
                1,
            ),
            District(
                "12th Essex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:12th_essex",
                1,
            ),
            District(
                "12th Hampden",
                "lower",
                "ocd-division/country:us/state:ma/sldl:12th_hampden",
                1,
            ),
            District(
                "12th Middlesex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:12th_middlesex",
                1,
            ),
            District(
                "12th Norfolk",
                "lower",
                "ocd-division/country:us/state:ma/sldl:12th_norfolk",
                1,
            ),
            District(
                "12th Plymouth",
                "lower",
                "ocd-division/country:us/state:ma/sldl:12th_plymouth",
                1,
            ),
            District(
                "12th Suffolk",
                "lower",
                "ocd-division/country:us/state:ma/sldl:12th_suffolk",
                1,
            ),
            District(
                "12th Worcester",
                "lower",
                "ocd-division/country:us/state:ma/sldl:12th_worcester",
                1,
            ),
            District(
                "13th Bristol",
                "lower",
                "ocd-division/country:us/state:ma/sldl:13th_bristol",
                1,
            ),
            District(
                "13th Essex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:13th_essex",
                1,
            ),
            District(
                "13th Middlesex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:13th_middlesex",
                1,
            ),
            District(
                "13th Norfolk",
                "lower",
                "ocd-division/country:us/state:ma/sldl:13th_norfolk",
                1,
            ),
            District(
                "13th Suffolk",
                "lower",
                "ocd-division/country:us/state:ma/sldl:13th_suffolk",
                1,
            ),
            District(
                "13th Worcester",
                "lower",
                "ocd-division/country:us/state:ma/sldl:13th_worcester",
                1,
            ),
            District(
                "14th Bristol",
                "lower",
                "ocd-division/country:us/state:ma/sldl:14th_bristol",
                1,
            ),
            District(
                "14th Essex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:14th_essex",
                1,
            ),
            District(
                "14th Middlesex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:14th_middlesex",
                1,
            ),
            District(
                "14th Norfolk",
                "lower",
                "ocd-division/country:us/state:ma/sldl:14th_norfolk",
                1,
            ),
            District(
                "14th Suffolk",
                "lower",
                "ocd-division/country:us/state:ma/sldl:14th_suffolk",
                1,
            ),
            District(
                "14th Worcester",
                "lower",
                "ocd-division/country:us/state:ma/sldl:14th_worcester",
                1,
            ),
            District(
                "15th Essex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:15th_essex",
                1,
            ),
            District(
                "15th Middlesex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:15th_middlesex",
                1,
            ),
            District(
                "15th Norfolk",
                "lower",
                "ocd-division/country:us/state:ma/sldl:15th_norfolk",
                1,
            ),
            District(
                "15th Suffolk",
                "lower",
                "ocd-division/country:us/state:ma/sldl:15th_suffolk",
                1,
            ),
            District(
                "15th Worcester",
                "lower",
                "ocd-division/country:us/state:ma/sldl:15th_worcester",
                1,
            ),
            District(
                "16th Essex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:16th_essex",
                1,
            ),
            District(
                "16th Middlesex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:16th_middlesex",
                1,
            ),
            District(
                "16th Suffolk",
                "lower",
                "ocd-division/country:us/state:ma/sldl:16th_suffolk",
                1,
            ),
            District(
                "16th Worcester",
                "lower",
                "ocd-division/country:us/state:ma/sldl:16th_worcester",
                1,
            ),
            District(
                "17th Essex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:17th_essex",
                1,
            ),
            District(
                "17th Middlesex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:17th_middlesex",
                1,
            ),
            District(
                "17th Suffolk",
                "lower",
                "ocd-division/country:us/state:ma/sldl:17th_suffolk",
                1,
            ),
            District(
                "17th Worcester",
                "lower",
                "ocd-division/country:us/state:ma/sldl:17th_worcester",
                1,
            ),
            District(
                "18th Essex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:18th_essex",
                1,
            ),
            District(
                "18th Middlesex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:18th_middlesex",
                1,
            ),
            District(
                "18th Suffolk",
                "lower",
                "ocd-division/country:us/state:ma/sldl:18th_suffolk",
                1,
            ),
            District(
                "19th Worcester",
                "lower",
                "ocd-division/country:us/state:ma/sldl:19th_worcester",
                1,
            ),
            District(
                "18th Worcester",
                "lower",
                "ocd-division/country:us/state:ma/sldl:18th_worcester",
                1,
            ),
            District(
                "19th Middlesex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:19th_middlesex",
                1,
            ),
            District(
                "19th Suffolk",
                "lower",
                "ocd-division/country:us/state:ma/sldl:19th_suffolk",
                1,
            ),
            District(
                "1st Barnstable",
                "lower",
                "ocd-division/country:us/state:ma/sldl:1st_barnstable",
                1,
            ),
            District(
                "1st Berkshire",
                "lower",
                "ocd-division/country:us/state:ma/sldl:1st_berkshire",
                1,
            ),
            District(
                "1st Bristol",
                "lower",
                "ocd-division/country:us/state:ma/sldl:1st_bristol",
                1,
            ),
            District(
                "1st Essex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:1st_essex",
                1,
            ),
            District(
                "1st Franklin",
                "lower",
                "ocd-division/country:us/state:ma/sldl:1st_franklin",
                1,
            ),
            District(
                "1st Hampden",
                "lower",
                "ocd-division/country:us/state:ma/sldl:1st_hampden",
                1,
            ),
            District(
                "1st Hampshire",
                "lower",
                "ocd-division/country:us/state:ma/sldl:1st_hampshire",
                1,
            ),
            District(
                "1st Middlesex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:1st_middlesex",
                1,
            ),
            District(
                "1st Norfolk",
                "lower",
                "ocd-division/country:us/state:ma/sldl:1st_norfolk",
                1,
            ),
            District(
                "1st Plymouth",
                "lower",
                "ocd-division/country:us/state:ma/sldl:1st_plymouth",
                1,
            ),
            District(
                "1st Suffolk",
                "lower",
                "ocd-division/country:us/state:ma/sldl:1st_suffolk",
                1,
            ),
            District(
                "1st Worcester",
                "lower",
                "ocd-division/country:us/state:ma/sldl:1st_worcester",
                1,
            ),
            District(
                "20th Middlesex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:20th_middlesex",
                1,
            ),
            District(
                "21st Middlesex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:21st_middlesex",
                1,
            ),
            District(
                "22nd Middlesex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:22nd_middlesex",
                1,
            ),
            District(
                "23rd Middlesex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:23rd_middlesex",
                1,
            ),
            District(
                "24th Middlesex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:24th_middlesex",
                1,
            ),
            District(
                "25th Middlesex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:25th_middlesex",
                1,
            ),
            District(
                "26th Middlesex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:26th_middlesex",
                1,
            ),
            District(
                "27th Middlesex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:27th_middlesex",
                1,
            ),
            District(
                "28th Middlesex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:28th_middlesex",
                1,
            ),
            District(
                "29th Middlesex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:29th_middlesex",
                1,
            ),
            District(
                "2nd Barnstable",
                "lower",
                "ocd-division/country:us/state:ma/sldl:2nd_barnstable",
                1,
            ),
            District(
                "2nd Berkshire",
                "lower",
                "ocd-division/country:us/state:ma/sldl:2nd_berkshire",
                1,
            ),
            District(
                "2nd Bristol",
                "lower",
                "ocd-division/country:us/state:ma/sldl:2nd_bristol",
                1,
            ),
            District(
                "2nd Essex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:2nd_essex",
                1,
            ),
            District(
                "2nd Franklin",
                "lower",
                "ocd-division/country:us/state:ma/sldl:2nd_franklin",
                1,
            ),
            District(
                "2nd Hampden",
                "lower",
                "ocd-division/country:us/state:ma/sldl:2nd_hampden",
                1,
            ),
            District(
                "2nd Hampshire",
                "lower",
                "ocd-division/country:us/state:ma/sldl:2nd_hampshire",
                1,
            ),
            District(
                "2nd Middlesex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:2nd_middlesex",
                1,
            ),
            District(
                "2nd Norfolk",
                "lower",
                "ocd-division/country:us/state:ma/sldl:2nd_norfolk",
                1,
            ),
            District(
                "2nd Plymouth",
                "lower",
                "ocd-division/country:us/state:ma/sldl:2nd_plymouth",
                1,
            ),
            District(
                "2nd Suffolk",
                "lower",
                "ocd-division/country:us/state:ma/sldl:2nd_suffolk",
                1,
            ),
            District(
                "2nd Worcester",
                "lower",
                "ocd-division/country:us/state:ma/sldl:2nd_worcester",
                1,
            ),
            District(
                "30th Middlesex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:30th_middlesex",
                1,
            ),
            District(
                "31st Middlesex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:31st_middlesex",
                1,
            ),
            District(
                "32nd Middlesex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:32nd_middlesex",
                1,
            ),
            District(
                "33rd Middlesex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:33rd_middlesex",
                1,
            ),
            District(
                "34th Middlesex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:34th_middlesex",
                1,
            ),
            District(
                "35th Middlesex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:35th_middlesex",
                1,
            ),
            District(
                "36th Middlesex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:36th_middlesex",
                1,
            ),
            District(
                "37th Middlesex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:37th_middlesex",
                1,
            ),
            District(
                "3rd Barnstable",
                "lower",
                "ocd-division/country:us/state:ma/sldl:3rd_barnstable",
                1,
            ),
            District(
                "3rd Berkshire",
                "lower",
                "ocd-division/country:us/state:ma/sldl:3rd_berkshire",
                1,
            ),
            District(
                "3rd Bristol",
                "lower",
                "ocd-division/country:us/state:ma/sldl:3rd_bristol",
                1,
            ),
            District(
                "3rd Essex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:3rd_essex",
                1,
            ),
            District(
                "3rd Hampden",
                "lower",
                "ocd-division/country:us/state:ma/sldl:3rd_hampden",
                1,
            ),
            District(
                "3rd Hampshire",
                "lower",
                "ocd-division/country:us/state:ma/sldl:3rd_hampshire",
                1,
            ),
            District(
                "3rd Middlesex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:3rd_middlesex",
                1,
            ),
            District(
                "3rd Norfolk",
                "lower",
                "ocd-division/country:us/state:ma/sldl:3rd_norfolk",
                1,
            ),
            District(
                "3rd Plymouth",
                "lower",
                "ocd-division/country:us/state:ma/sldl:3rd_plymouth",
                1,
            ),
            District(
                "3rd Suffolk",
                "lower",
                "ocd-division/country:us/state:ma/sldl:3rd_suffolk",
                1,
            ),
            District(
                "3rd Worcester",
                "lower",
                "ocd-division/country:us/state:ma/sldl:3rd_worcester",
                1,
            ),
            District(
                "4th Barnstable",
                "lower",
                "ocd-division/country:us/state:ma/sldl:4th_barnstable",
                1,
            ),
            District(
                "4th Berkshire",
                "lower",
                "ocd-division/country:us/state:ma/sldl:4th_berkshire",
                1,
            ),
            District(
                "4th Bristol",
                "lower",
                "ocd-division/country:us/state:ma/sldl:4th_bristol",
                1,
            ),
            District(
                "4th Essex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:4th_essex",
                1,
            ),
            District(
                "4th Hampden",
                "lower",
                "ocd-division/country:us/state:ma/sldl:4th_hampden",
                1,
            ),
            District(
                "4th Middlesex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:4th_middlesex",
                1,
            ),
            District(
                "4th Norfolk",
                "lower",
                "ocd-division/country:us/state:ma/sldl:4th_norfolk",
                1,
            ),
            District(
                "4th Plymouth",
                "lower",
                "ocd-division/country:us/state:ma/sldl:4th_plymouth",
                1,
            ),
            District(
                "4th Suffolk",
                "lower",
                "ocd-division/country:us/state:ma/sldl:4th_suffolk",
                1,
            ),
            District(
                "4th Worcester",
                "lower",
                "ocd-division/country:us/state:ma/sldl:4th_worcester",
                1,
            ),
            District(
                "5th Barnstable",
                "lower",
                "ocd-division/country:us/state:ma/sldl:5th_barnstable",
                1,
            ),
            District(
                "5th Bristol",
                "lower",
                "ocd-division/country:us/state:ma/sldl:5th_bristol",
                1,
            ),
            District(
                "5th Essex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:5th_essex",
                1,
            ),
            District(
                "5th Hampden",
                "lower",
                "ocd-division/country:us/state:ma/sldl:5th_hampden",
                1,
            ),
            District(
                "5th Middlesex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:5th_middlesex",
                1,
            ),
            District(
                "5th Norfolk",
                "lower",
                "ocd-division/country:us/state:ma/sldl:5th_norfolk",
                1,
            ),
            District(
                "5th Plymouth",
                "lower",
                "ocd-division/country:us/state:ma/sldl:5th_plymouth",
                1,
            ),
            District(
                "5th Suffolk",
                "lower",
                "ocd-division/country:us/state:ma/sldl:5th_suffolk",
                1,
            ),
            District(
                "5th Worcester",
                "lower",
                "ocd-division/country:us/state:ma/sldl:5th_worcester",
                1,
            ),
            District(
                "6th Bristol",
                "lower",
                "ocd-division/country:us/state:ma/sldl:6th_bristol",
                1,
            ),
            District(
                "6th Essex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:6th_essex",
                1,
            ),
            District(
                "6th Hampden",
                "lower",
                "ocd-division/country:us/state:ma/sldl:6th_hampden",
                1,
            ),
            District(
                "6th Middlesex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:6th_middlesex",
                1,
            ),
            District(
                "6th Norfolk",
                "lower",
                "ocd-division/country:us/state:ma/sldl:6th_norfolk",
                1,
            ),
            District(
                "6th Plymouth",
                "lower",
                "ocd-division/country:us/state:ma/sldl:6th_plymouth",
                1,
            ),
            District(
                "6th Suffolk",
                "lower",
                "ocd-division/country:us/state:ma/sldl:6th_suffolk",
                1,
            ),
            District(
                "6th Worcester",
                "lower",
                "ocd-division/country:us/state:ma/sldl:6th_worcester",
                1,
            ),
            District(
                "7th Bristol",
                "lower",
                "ocd-division/country:us/state:ma/sldl:7th_bristol",
                1,
            ),
            District(
                "7th Essex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:7th_essex",
                1,
            ),
            District(
                "7th Hampden",
                "lower",
                "ocd-division/country:us/state:ma/sldl:7th_hampden",
                1,
            ),
            District(
                "7th Middlesex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:7th_middlesex",
                1,
            ),
            District(
                "7th Norfolk",
                "lower",
                "ocd-division/country:us/state:ma/sldl:7th_norfolk",
                1,
            ),
            District(
                "7th Plymouth",
                "lower",
                "ocd-division/country:us/state:ma/sldl:7th_plymouth",
                1,
            ),
            District(
                "7th Suffolk",
                "lower",
                "ocd-division/country:us/state:ma/sldl:7th_suffolk",
                1,
            ),
            District(
                "7th Worcester",
                "lower",
                "ocd-division/country:us/state:ma/sldl:7th_worcester",
                1,
            ),
            District(
                "8th Bristol",
                "lower",
                "ocd-division/country:us/state:ma/sldl:8th_bristol",
                1,
            ),
            District(
                "8th Essex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:8th_essex",
                1,
            ),
            District(
                "8th Hampden",
                "lower",
                "ocd-division/country:us/state:ma/sldl:8th_hampden",
                1,
            ),
            District(
                "8th Middlesex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:8th_middlesex",
                1,
            ),
            District(
                "8th Norfolk",
                "lower",
                "ocd-division/country:us/state:ma/sldl:8th_norfolk",
                1,
            ),
            District(
                "8th Plymouth",
                "lower",
                "ocd-division/country:us/state:ma/sldl:8th_plymouth",
                1,
            ),
            District(
                "8th Suffolk",
                "lower",
                "ocd-division/country:us/state:ma/sldl:8th_suffolk",
                1,
            ),
            District(
                "8th Worcester",
                "lower",
                "ocd-division/country:us/state:ma/sldl:8th_worcester",
                1,
            ),
            District(
                "9th Bristol",
                "lower",
                "ocd-division/country:us/state:ma/sldl:9th_bristol",
                1,
            ),
            District(
                "18th Essex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:18th_essex",
                1,
            ),
            District(
                "9th Essex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:9th_essex",
                1,
            ),
            District(
                "9th Hampden",
                "lower",
                "ocd-division/country:us/state:ma/sldl:9th_hampden",
                1,
            ),
            District(
                "9th Middlesex",
                "lower",
                "ocd-division/country:us/state:ma/sldl:9th_middlesex",
                1,
            ),
            District(
                "9th Norfolk",
                "lower",
                "ocd-division/country:us/state:ma/sldl:9th_norfolk",
                1,
            ),
            District(
                "9th Plymouth",
                "lower",
                "ocd-division/country:us/state:ma/sldl:9th_plymouth",
                1,
            ),
            District(
                "9th Suffolk",
                "lower",
                "ocd-division/country:us/state:ma/sldl:9th_suffolk",
                1,
            ),
            District(
                "9th Worcester",
                "lower",
                "ocd-division/country:us/state:ma/sldl:9th_worcester",
                1,
            ),
            District(
                "Barnstable, Dukes and Nantucket",
                "lower",
                "ocd-division/country:us/state:ma/sldl:barnstable_dukes_and_nantucket",
                1,
            ),
        ],
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/1a75ab3a-669b-43fe-ac8d-31a2d6923d9a",
        num_seats=40,
        title="Senator",
        districts=[
            District(
                "Bristol and Norfolk",
                "upper",
                "ocd-division/country:us/state:ma/sldu:bristol_and_norfolk",
                1,
            ),
            District(
                "Cape and Islands",
                "upper",
                "ocd-division/country:us/state:ma/sldu:cape_and_islands",
                1,
            ),
            District(
                "Fifth Middlesex",
                "upper",
                "ocd-division/country:us/state:ma/sldu:5th_middlesex",
                1,
            ),
            District(
                "First Bristol and Plymouth",
                "upper",
                "ocd-division/country:us/state:ma/sldu:1st_bristol_and_plymouth",
                1,
            ),
            District(
                "First Essex",
                "upper",
                "ocd-division/country:us/state:ma/sldu:1st_essex",
                1,
            ),
            District(
                "First Essex and Middlesex",
                "upper",
                "ocd-division/country:us/state:ma/sldu:1st_essex_and_middlesex",
                1,
            ),
            District(
                "First Hampden and Hampshire",
                "upper",
                "ocd-division/country:us/state:ma/sldu:1st_hampden_and_hampshire",
                1,
            ),
            District(
                "First Middlesex",
                "upper",
                "ocd-division/country:us/state:ma/sldu:1st_middlesex",
                1,
            ),
            District(
                "First Middlesex and Norfolk",
                "upper",
                "ocd-division/country:us/state:ma/sldu:1st_middlesex_and_norfolk",
                1,
            ),
            District(
                "First Plymouth and Bristol",
                "upper",
                "ocd-division/country:us/state:ma/sldu:1st_plymouth_and_bristol",
                1,
            ),
            District(
                "First Suffolk",
                "upper",
                "ocd-division/country:us/state:ma/sldu:1st_suffolk",
                1,
            ),
            District(
                "First Suffolk and Middlesex",
                "upper",
                "ocd-division/country:us/state:ma/sldu:1st_suffolk_and_middlesex",
                1,
            ),
            District(
                "First Worcester",
                "upper",
                "ocd-division/country:us/state:ma/sldu:1st_worcester",
                1,
            ),
            District(
                "Fourth Middlesex",
                "upper",
                "ocd-division/country:us/state:ma/sldu:4th_middlesex",
                1,
            ),
            District(
                "Hampden", "upper", "ocd-division/country:us/state:ma/sldu:hampden", 1
            ),
            District(
                "Berkshire, Hampshire, Franklin and Hampden",
                "upper",
                "ocd-division/country:us/state:ma/sldu:berkshire_hampshire_franklin_and_hampden",
                1,
            ),
            District(
                "Middlesex and Suffolk",
                "upper",
                "ocd-division/country:us/state:ma/sldu:middlesex_and_suffolk",
                1,
            ),
            District(
                "Middlesex and Worcester",
                "upper",
                "ocd-division/country:us/state:ma/sldu:middlesex_and_worcester",
                1,
            ),
            District(
                "Norfolk, Bristol and Middlesex",
                "upper",
                "ocd-division/country:us/state:ma/sldu:norfolk_bristol_and_middlesex",
                1,
            ),
            District(
                "Norfolk, Bristol and Plymouth",
                "upper",
                "ocd-division/country:us/state:ma/sldu:norfolk_bristol_and_plymouth",
                1,
            ),
            District(
                "Norfolk and Plymouth",
                "upper",
                "ocd-division/country:us/state:ma/sldu:norfolk_and_plymouth",
                1,
            ),
            District(
                "Norfolk and Suffolk",
                "upper",
                "ocd-division/country:us/state:ma/sldu:norfolk_and_suffolk",
                1,
            ),
            District(
                "Plymouth and Barnstable",
                "upper",
                "ocd-division/country:us/state:ma/sldu:plymouth_and_barnstable",
                1,
            ),
            District(
                "Plymouth and Norfolk",
                "upper",
                "ocd-division/country:us/state:ma/sldu:plymouth_and_norfolk",
                1,
            ),
            District(
                "Second Bristol and Plymouth",
                "upper",
                "ocd-division/country:us/state:ma/sldu:2nd_bristol_and_plymouth",
                1,
            ),
            District(
                "Second Essex",
                "upper",
                "ocd-division/country:us/state:ma/sldu:2nd_essex",
                1,
            ),
            District(
                "Second Essex and Middlesex",
                "upper",
                "ocd-division/country:us/state:ma/sldu:2nd_essex_and_middlesex",
                1,
            ),
            District(
                "Second Hampden and Hampshire",
                "upper",
                "ocd-division/country:us/state:ma/sldu:2nd_hampden_and_hampshire",
                1,
            ),
            District(
                "Second Middlesex",
                "upper",
                "ocd-division/country:us/state:ma/sldu:2nd_middlesex",
                1,
            ),
            District(
                "Second Middlesex and Norfolk",
                "upper",
                "ocd-division/country:us/state:ma/sldu:2nd_middlesex_and_norfolk",
                1,
            ),
            District(
                "Second Plymouth and Bristol",
                "upper",
                "ocd-division/country:us/state:ma/sldu:2nd_plymouth_and_bristol",
                1,
            ),
            District(
                "Second Suffolk",
                "upper",
                "ocd-division/country:us/state:ma/sldu:2nd_suffolk",
                1,
            ),
            District(
                "Second Suffolk and Middlesex",
                "upper",
                "ocd-division/country:us/state:ma/sldu:2nd_suffolk_and_middlesex",
                1,
            ),
            District(
                "Second Worcester",
                "upper",
                "ocd-division/country:us/state:ma/sldu:2nd_worcester",
                1,
            ),
            District(
                "Third Essex",
                "upper",
                "ocd-division/country:us/state:ma/sldu:3rd_essex",
                1,
            ),
            District(
                "Third Middlesex",
                "upper",
                "ocd-division/country:us/state:ma/sldu:3rd_middlesex",
                1,
            ),
            District(
                "Worcester and Middlesex",
                "upper",
                "ocd-division/country:us/state:ma/sldu:worcester_and_middlesex",
                1,
            ),
            District(
                "Worcester and Norfolk",
                "upper",
                "ocd-division/country:us/state:ma/sldu:worcester_and_norfolk",
                1,
            ),
            District(
                "Worcester, Hampden, Hampshire and Middlesex",
                "upper",
                "ocd-division/country:us/state:ma/sldu:worcester_hampden_hampshire_and_middlesex",
                1,
            ),
            District(
                "Hampshire, Franklin and Worcester",
                "upper",
                "ocd-division/country:us/state:ma/sldu:hampshire_franklin_and_worcester",
                1,
            ),
        ],
    ),
)



import pytest  # type: ignore
import django  # type: ignore
from openstates.data.models import (
    Jurisdiction,
    Division,
    Organization,
    Person,
    Bill,
    VoteEvent,
)
from openstates.cli.reports import generate_session_report


def create_data():
    Division.objects.create(id="ocd-division/country:us", name="USA")
    j = Jurisdiction.objects.create(id="jid", division_id="ocd-division/country:us")
    org = Organization.objects.create(
        jurisdiction=j, name="House", classification="lower"
    )
    person = Person.objects.create(name="Roy")
    j.legislative_sessions.create(identifier="1899", name="1899")
    session = j.legislative_sessions.create(identifier="1900", name="1900").id
    return session, org, person


@pytest.mark.django_db
def test_bills_missing_actions():
    session, org, person = create_data()
    Bill.objects.create(identifier="HB1", title="One", legislative_session_id=session)
    b = Bill.objects.create(
        identifier="HB2", title="Two", legislative_session_id=session
    )

    report = generate_session_report(session)
    assert report.bills_missing_actions == 2

    b.actions.create(description="Introduced", order=1, organization=org)
    report = generate_session_report(session)
    assert report.bills_missing_actions == 1


@pytest.mark.django_db
def test_bills_missing_sponsors():
    session, org, person = create_data()
    Bill.objects.create(identifier="HB1", title="One", legislative_session_id=session)
    b = Bill.objects.create(
        identifier="HB2", title="Two", legislative_session_id=session
    )

    report = generate_session_report(session)
    assert report.bills_missing_sponsors == 2

    b.sponsorships.create(name="Roy", entity_type="person")
    report = generate_session_report(session)
    assert report.bills_missing_sponsors == 1


@pytest.mark.django_db
def test_bills_missing_versions():
    session, org, person = create_data()
    Bill.objects.create(identifier="HB1", title="One", legislative_session_id=session)
    b = Bill.objects.create(
        identifier="HB2", title="Two", legislative_session_id=session
    )

    report = generate_session_report(session)
    assert report.bills_missing_versions == 2

    b.versions.create(note="Final Copy")
    report = generate_session_report(session)
    assert report.bills_missing_versions == 1


@pytest.mark.django_db
def test_votes_missing_bill():
    session, org, person = create_data()
    b = Bill.objects.create(
        identifier="HB2", title="Two", legislative_session_id=session
    )
    v = VoteEvent.objects.create(
        legislative_session_id=session, motion_text="Passage", organization=org
    )
    VoteEvent.objects.create(
        legislative_session_id=session, motion_text="Amendment", organization=org
    )

    report = generate_session_report(session)
    assert report.votes_missing_bill == 2

    v.bill = b
    v.save()
    report = generate_session_report(session)
    assert report.votes_missing_bill == 1


@pytest.mark.django_db
def test_votes_missing_voters():
    session, org, person = create_data()
    b = Bill.objects.create(
        identifier="HB2", title="Two", legislative_session_id=session
    )
    v = VoteEvent.objects.create(
        legislative_session_id=session, motion_text="Passage", bill=b, organization=org
    )
    VoteEvent.objects.create(
        legislative_session_id=session,
        motion_text="Amendment",
        bill=b,
        organization=org,
    )

    report = generate_session_report(session)
    assert report.votes_missing_voters == 2

    v.votes.create(option="yes", voter_name="Speaker")
    report = generate_session_report(session)
    assert report.votes_missing_voters == 1


@pytest.mark.django_db
def test_missing_yes_no_counts():
    session, org, person = create_data()
    b = Bill.objects.create(
        identifier="HB2", title="Two", legislative_session_id=session
    )
    v = VoteEvent.objects.create(
        legislative_session_id=session, motion_text="Passage", bill=b, organization=org
    )
    VoteEvent.objects.create(
        legislative_session_id=session,
        motion_text="Amendment",
        bill=b,
        organization=org,
    )

    report = generate_session_report(session)
    assert report.votes_missing_yes_count == 2
    assert report.votes_missing_no_count == 2

    v.counts.create(option="yes", value=1)
    report = generate_session_report(session)
    assert report.votes_missing_yes_count == 1
    assert report.votes_missing_no_count == 2

    v.counts.create(option="no", value=0)
    report = generate_session_report(session)
    assert report.votes_missing_yes_count == 1
    assert report.votes_missing_no_count == 1


@pytest.mark.skipif(django.VERSION[0] < 2, reason="bug in Django 1.11")
@pytest.mark.django_db
def test_votes_with_bad_counts():
    session, org, person = create_data()
    b = Bill.objects.create(
        identifier="HB2", title="Two", legislative_session_id=session
    )
    v = VoteEvent.objects.create(
        legislative_session_id=session, motion_text="Passage", bill=b, organization=org
    )

    report = generate_session_report(session)
    assert report.votes_with_bad_counts == 0

    # add count, breaking
    v.counts.create(option="yes", value=1)
    report = generate_session_report(session)
    assert report.votes_with_bad_counts == 1

    # add voter, fixing
    v.votes.create(option="yes", voter_name="One")
    report = generate_session_report(session)
    assert report.votes_with_bad_counts == 0

    # add voter, breaking
    v.votes.create(option="no", voter_name="Two")
    report = generate_session_report(session)
    assert report.votes_with_bad_counts == 1

    # add count, still not equal
    v.counts.create(option="no", value=2)
    report = generate_session_report(session)
    assert report.votes_with_bad_counts == 1

    # add voter, fixing
    v.votes.create(option="no", voter_name="Three")
    report = generate_session_report(session)
    assert report.votes_with_bad_counts == 0


@pytest.mark.django_db
def test_unmatched_sponsors():
    session, org, person = create_data()
    b1 = Bill.objects.create(
        identifier="HB1", title="One", legislative_session_id=session
    )
    b2 = Bill.objects.create(
        identifier="HB2", title="Two", legislative_session_id=session
    )

    b1.sponsorships.create(name="Roy", entity_type="person")
    b1.sponsorships.create(name="Wendy", entity_type="person")
    b1.sponsorships.create(name="Committee On Legislation", entity_type="organization")

    b2.sponsorships.create(name="Wendy", entity_type="person")

    report = generate_session_report(session)
    assert len(report.unmatched_sponsor_people) == 2
    assert report.unmatched_sponsor_people["Roy"] == 1
    assert report.unmatched_sponsor_people["Wendy"] == 2
    assert report.unmatched_sponsor_organizations == {"Committee On Legislation": 1}

    # ensure that Roy goes away when linked
    sp = b1.sponsorships.get(name="Roy")
    sp.person_id = person.id
    sp.save()
    report = generate_session_report(session)
    assert report.unmatched_sponsor_people == {"Wendy": 2}


@pytest.mark.django_db
def test_unmatched_voters():
    session, org, person = create_data()
    b = Bill.objects.create(
        identifier="HB2", title="Two", legislative_session_id=session
    )
    v1 = VoteEvent.objects.create(
        legislative_session_id=session, motion_text="Passage", bill=b, organization=org
    )
    v2 = VoteEvent.objects.create(
        legislative_session_id=session, motion_text="Override", bill=b, organization=org
    )

    report = generate_session_report(session)
    assert report.unmatched_voters == {}

    # add voters
    v1.votes.create(option="yes", voter_name="Roy")
    v1.votes.create(option="yes", voter_name="Wendy")
    v2.votes.create(option="yes", voter_name="Wendy")
    report = generate_session_report(session)
    assert len(report.unmatched_voters) == 2
    assert report.unmatched_voters["Roy"] == 1
    assert report.unmatched_voters["Wendy"] == 2

    # ensure that Roy goes away when linked
    voter = v1.votes.get(voter_name="Roy")
    voter.voter_id = person.id
    voter.save()
    report = generate_session_report(session)
    assert report.unmatched_voters == {"Wendy": 2}



# Generated by Django 3.0.5 on 2020-04-22 00:28

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("data", "0001_initial")]

    operations = [
        migrations.RemoveField(
            model_name="organizationidentifier", name="organization"
        ),
        migrations.RemoveField(model_name="organizationname", name="organization"),
        migrations.RemoveField(model_name="postcontactdetail", name="post"),
        migrations.RemoveField(model_name="postlink", name="post"),
        migrations.DeleteModel(name="OrganizationContactDetail"),
        migrations.DeleteModel(name="OrganizationIdentifier"),
        migrations.DeleteModel(name="OrganizationName"),
        migrations.DeleteModel(name="PostContactDetail"),
        migrations.DeleteModel(name="PostLink"),
    ]



# Generated by Django 3.0.5 on 2020-09-14 10:36

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("data", "0018_auto_20200914_1032"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="billabstract",
            name="date",
        ),
    ]



import os
import json
import shutil
import tempfile
import datetime
import pytest
from unittest import mock
from openstates.data.models import (
    Bill,
    Jurisdiction,
    Division,
    LegislativeSession,
    Organization,
    Person,
)
from openstates.scrape import Bill as ScrapeBill
from openstates.importers.base import omnihash, BaseImporter
from openstates.importers import BillImporter
from openstates.exceptions import UnresolvedIdError, DataImportError


def create_jurisdiction():
    Division.objects.create(id="ocd-division/country:us", name="USA")
    j = Jurisdiction.objects.create(id="jid", division_id="ocd-division/country:us")
    Organization.objects.create(jurisdiction_id="jid", classification="legislature")
    LegislativeSession.objects.create(
        jurisdiction_id="jid", name="2020", identifier="2020"
    )
    return j


class FakeImporter(BaseImporter):
    _type = "test"


def test_omnihash_python_types():
    # string
    assert omnihash("test") == omnihash("test")
    # list
    assert omnihash(["this", "is", "a", "list"]) == omnihash(
        ["this", "is", "a", "list"]
    )
    # set
    assert omnihash({"and", "a", "set"}) == omnihash({"set", "set", "and", "a"})
    # dict w/ set and tuple as well
    assert omnihash({"a": {("fancy", "nested"): {"dict"}}}) == omnihash(
        {"a": {("fancy", "nested"): {"dict"}}}
    )


def test_import_directory():
    # write out some temp data to filesystem
    datadir = tempfile.mkdtemp()
    dicta = {"test": "A"}
    dictb = {"test": "B"}
    open(os.path.join(datadir, "test_a.json"), "w").write(json.dumps(dicta))
    open(os.path.join(datadir, "test_b.json"), "w").write(json.dumps(dictb))

    # simply ensure that import directory calls import_data with all dicts
    ti = FakeImporter("jurisdiction-id")
    with mock.patch.object(ti, attribute="import_data") as mockobj:
        ti.import_directory(datadir)

    # import_data should be called once
    assert mockobj.call_count == 1
    # kind of hacky, get the total list of args passed in
    arg_objs = list(mockobj.call_args[0][0])

    # 2 args only, make sure a and b are in there
    assert len(arg_objs) == 2
    assert dicta in arg_objs
    assert dictb in arg_objs

    # clean up datadir
    shutil.rmtree(datadir)


def test_apply_transformers():
    transformers = {
        "capitalize": lambda x: x.upper(),
        "cap_and_reverse": [lambda x: x.upper(), lambda y: y[::-1]],
        "never_used": lambda x: 1 / 0,
        "nested": {"replace": lambda x: "replaced"},
    }
    data = {
        "capitalize": "words",
        "cap_and_reverse": "simple",
        "nested": {"replace": None},
    }
    ti = FakeImporter("jid")
    ti.cached_transformers = transformers
    output = ti.apply_transformers(data)
    assert output["capitalize"] == "WORDS"
    assert output["cap_and_reverse"] == "ELPMIS"
    assert output["nested"]["replace"] == "replaced"


# doing these next few tests just on a Bill because it is the same code that handles it
# but for completeness maybe it is better to do these on each type?


@pytest.mark.django_db
def test_deduplication_identical_object():
    create_jurisdiction()
    p1 = ScrapeBill("HB 1", "2020", "Title").as_dict()
    p2 = ScrapeBill("HB 1", "2020", "Title").as_dict()
    BillImporter("jid").import_data([p1, p2])

    assert Bill.objects.count() == 1


@pytest.mark.django_db
def test_exception_on_identical_objects_in_import_stream():
    create_jurisdiction()
    # the first two objects aren't identical, but refer to the same thing
    # at the moment we drop both objects (because we can't know which one is correct)
    b1 = ScrapeBill("HB 1", "2020", "Title", chamber="upper").as_dict()
    b2 = ScrapeBill("HB 1", "2020", "Title", chamber="lower").as_dict()
    b3 = ScrapeBill("HB 2", "2020", "Bill Title").as_dict()

    BillImporter("jid").import_data([b1, b2, b3])
    assert Bill.objects.count() == 1


@pytest.mark.django_db
def test_resolve_json_id():
    create_jurisdiction()
    p1 = ScrapeBill("HB 1", "2020", "Title").as_dict()
    p2 = ScrapeBill("HB 1", "2020", "Title").as_dict()
    bi = BillImporter("jid")

    # do import and get database id
    p1_id = p1["_id"]
    p2_id = p2["_id"]
    bi.import_data([p1, p2])
    db_id = Bill.objects.get().id

    # simplest case
    assert bi.resolve_json_id(p1_id) == db_id
    # duplicate should resolve to same id
    assert bi.resolve_json_id(p2_id) == db_id
    # a null id should map to None
    assert bi.resolve_json_id(None) is None
    # no such id
    with pytest.raises(UnresolvedIdError):
        bi.resolve_json_id("this-is-invalid")


@pytest.mark.django_db
def test_invalid_fields():
    create_jurisdiction()
    p1 = ScrapeBill("HB 1", "2020", "Title").as_dict()
    p1["newfield"] = "shouldn't happen"

    with pytest.raises(DataImportError):
        BillImporter("jid").import_data([p1])


@pytest.mark.django_db
def test_invalid_fields_related_item():
    create_jurisdiction()
    p1 = ScrapeBill("HB 1", "2020", "Title")
    p1.add_source("http://example.com")
    p1 = p1.as_dict()
    p1["sources"][0]["test"] = 3

    with pytest.raises(DataImportError):
        BillImporter("jid").import_data([p1])


@pytest.mark.django_db
def test_automatic_updated_at():
    create_jurisdiction()
    difference = Organization.objects.get().updated_at - datetime.datetime.utcnow()
    # updated_at should be in UTC, a bit of clock drift notwithstanding
    assert abs(difference) < datetime.timedelta(minutes=5)


@pytest.mark.django_db
def test_resolve_person_normal():
    create_jurisdiction()
    bi = BillImporter("jid")
    org = Organization.objects.get(jurisdiction_id="jid", classification="legislature")
    p = Person.objects.create(name="John McGuirk")
    p.memberships.create(organization=org)

    assert bi.resolve_person('~{"name": "John McGuirk"}') == p.id


@pytest.mark.django_db
def test_resolve_person_case_insensitive():
    create_jurisdiction()
    bi = BillImporter("jid")
    org = Organization.objects.get(jurisdiction_id="jid", classification="legislature")
    p = Person.objects.create(name="John McGuirk")
    p.memberships.create(organization=org)

    assert bi.resolve_person('~{"name": "JohN mCgUIrk"}') == p.id


@pytest.mark.django_db
def test_resolve_bill_by_date():
    j = create_jurisdiction()
    session = j.legislative_sessions.create(
        name="2021",
        identifier="2021",
        start_date="2021-01-01",
        end_date="2021-12-31",
    )
    bi = BillImporter("jid")
    b = Bill.objects.create(
        identifier="HB 1", title="Some Bill", legislative_session=session
    )

    assert bi.resolve_bill("HB 1", date="2021-05-06") == b.id


@pytest.mark.django_db
def test_resolve_bill_by_date_transformers():
    j = create_jurisdiction()
    session = j.legislative_sessions.create(
        name="2021",
        identifier="2021",
        start_date="2021-01-01",
        end_date="2021-12-31",
    )
    bi = BillImporter("jid")
    b = Bill.objects.create(
        identifier="HB 1", title="Some Bill", legislative_session=session
    )
    assert bi.resolve_bill("hb1", date="2021-05-06") == b.id


@pytest.mark.django_db
def test_resolve_bill_by_date_open_session():
    j = create_jurisdiction()
    session = j.legislative_sessions.create(
        name="2021",
        identifier="2021",
        start_date="2021-01-01",
        # no end date
    )
    bi = BillImporter("jid")
    b = Bill.objects.create(
        identifier="HB 1", title="Some Bill", legislative_session=session
    )

    assert bi.resolve_bill("HB 1", date="2021-05-06") == b.id



# Generated by Django 3.0.5 on 2020-04-28 15:29

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("data", "0007_auto_20200423_1416")]

    operations = [
        migrations.AddField(
            model_name="person",
            name="current_state",
            field=models.CharField(default="", max_length=2),
        )
    ]



from ..models import State, Chamber, simple_numbered_districts

UT = State(
    name="Utah",
    abbr="UT",
    capital="Salt Lake City",
    capital_tz="America/Denver",
    fips="49",
    unicameral=False,
    legislature_name="Utah State Legislature",
    legislature_organization_id="ocd-organization/58fd9e6c-78ea-42ff-b58b-4f80d307cf19",
    executive_name="Office of the Governor",
    executive_organization_id="ocd-organization/75532e47-e012-442d-b2de-f650c124c0b9",
    division_id="ocd-division/country:us/state:ut",
    jurisdiction_id="ocd-jurisdiction/country:us/state:ut/government",
    url="http://le.utah.gov/",
    lower=Chamber(
        chamber_type="lower",
        name="House",
        organization_id="ocd-organization/c7625e7e-dc12-49cf-aa73-fa96c5bf4b11",
        num_seats=75,
        title="Representative",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:ut", "lower", 75
        ),
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/98cb21d7-f3ae-416b-93e7-df3da962aff9",
        num_seats=29,
        title="Senator",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:ut", "upper", 29
        ),
    ),
)



import argparse
import contextlib
import datetime
import glob
import importlib
import inspect
import logging
import logging.config
import os
import sys
import traceback
import typing
from collections import defaultdict
from types import ModuleType

from django.db import transaction  # type: ignore

from .. import settings, utils
from ..exceptions import CommandError
from ..scrape import JurisdictionScraper, State
from ..utils.django import init_django
from ..utils.instrument import Instrumentation
from .reports import generate_session_report, print_report, save_report

logger = logging.getLogger("openstates")
stats = Instrumentation()

ALL_ACTIONS = ("scrape", "import")


class _Unset:
    pass


UNSET = _Unset()


@contextlib.contextmanager
def override_settings(settings, overrides):  # type: ignore
    original = {}
    for key, value in overrides.items():
        original[key] = getattr(settings, key, UNSET)
        setattr(settings, key, value)
    yield
    for key, value in original.items():
        if value is UNSET:
            delattr(settings, key)
        else:
            setattr(settings, key, value)


def get_jurisdiction(module_name: str) -> tuple[State, ModuleType]:
    # get the state object
    module = importlib.import_module(module_name)
    for obj in module.__dict__.values():
        # ensure we're dealing with a subclass of State
        if isinstance(obj, type) and issubclass(obj, State) and obj != State:
            return obj(), module
    raise CommandError(f"Unable to import State subclass from {module_name}")


def do_scrape(
    juris: State,
    args: argparse.Namespace,
    scrapers: dict[str, dict[str, str]],
    active_sessions: set[str],
) -> dict[str, typing.Any]:
    # make output and cache dirs
    utils.makedirs(settings.CACHE_DIR)
    datadir = os.path.join(settings.SCRAPED_DATA_DIR, args.module)
    utils.makedirs(datadir)
    # clear json from data dir
    for f in glob.glob(datadir + "/*.json"):
        os.remove(f)

    report = {}

    # do jurisdiction
    jscraper = JurisdictionScraper(
        juris,
        datadir,
        strict_validation=args.strict,
        fastmode=args.fastmode,
        realtime=args.realtime,
    )
    report["jurisdiction"] = jscraper.do_scrape()
    stats.send_counter("jurisdiction_scrapes_total", 1, {"jurisdiction": juris.name})

    for scraper_name, scrape_args in scrapers.items():
        ScraperCls = juris.scrapers[scraper_name]
        if (
            "session" in inspect.getargspec(ScraperCls.scrape).args
            and "session" not in scrape_args
        ):
            logger.warning(
                f"no session provided, using active sessions: {active_sessions}"
            )
            # handle automatically setting session if required by the scraper
            # the report logic was originally meant for one run, so we combine the start & end times
            # and counts here
            report[scraper_name] = {
                "start": None,
                "end": None,
                "objects": defaultdict(int),
            }
            for session in active_sessions:
                # new scraper each time
                scraper = ScraperCls(
                    juris,
                    datadir,
                    strict_validation=args.strict,
                    fastmode=args.fastmode,
                    realtime=args.realtime,
                )
                partial_report = scraper.do_scrape(**scrape_args, session=session)
                stats.send_counter(
                    "session_scrapes_total",
                    1,
                    {"jurisdiction": juris.name, "session": session},
                )
                if not report[scraper_name]["start"]:
                    report[scraper_name]["start"] = partial_report["start"]
                report[scraper_name]["end"] = partial_report["end"]
                for obj, val in partial_report["objects"].items():
                    report[scraper_name]["objects"][obj] += val
                stats.send_last_run(
                    "last_session_scrape_time",
                    {"jurisdiction": juris.name, "session": session},
                )
        else:
            scraper = ScraperCls(
                juris,
                datadir,
                strict_validation=args.strict,
                fastmode=args.fastmode,
                realtime=args.realtime,
            )
            report[scraper_name] = scraper.do_scrape(**scrape_args)
            session = scrape_args.get("session", "")
            if session:
                stats.send_counter(
                    "session_scrapes_total",
                    1,
                    {"jurisdiction": juris.name, "session": session},
                )
                stats.send_last_run(
                    "last_session_scrape_time",
                    {"jurisdiction": juris.name, "session": session},
                )
            else:
                stats.send_counter(
                    "non_session_scrapes_total",
                    1,
                    {
                        "jurisdiction": juris.name,
                    },
                )
                stats.send_last_run(
                    "last_non_session_scrape_time",
                    {
                        "jurisdiction": juris.name,
                    },
                )

    return report


def do_import(juris: State, args: argparse.Namespace) -> dict[str, typing.Any]:
    # import inside here because to avoid loading Django code unnecessarily
    from openstates.data.models import Jurisdiction as DatabaseJurisdiction
    from openstates.importers import (
        BillImporter,
        EventImporter,
        JurisdictionImporter,
        VoteEventImporter,
    )

    datadir = os.path.join(settings.SCRAPED_DATA_DIR, args.module)

    juris_importer = JurisdictionImporter(juris.jurisdiction_id)
    bill_importer = BillImporter(juris.jurisdiction_id)
    vote_event_importer = VoteEventImporter(juris.jurisdiction_id, bill_importer)
    event_importer = EventImporter(juris.jurisdiction_id, vote_event_importer)
    report = {}

    with transaction.atomic():
        logger.info("import jurisdictions...")
        report.update(juris_importer.import_directory(datadir))
        logger.info("import bills...")
        report.update(bill_importer.import_directory(datadir))
        logger.info("import vote events...")
        report.update(vote_event_importer.import_directory(datadir))
        logger.info("import events...")
        report.update(event_importer.import_directory(datadir))
        DatabaseJurisdiction.objects.filter(id=juris.jurisdiction_id).update(
            latest_bill_update=datetime.datetime.utcnow()
        )

    # compile info on all sessions that were updated in this run
    seen_sessions = set()
    seen_sessions.update(bill_importer.get_seen_sessions())
    seen_sessions.update(vote_event_importer.get_seen_sessions())
    for session in seen_sessions:
        generate_session_report(session)

    return report


def check_session_list(juris: State) -> set[str]:
    scraper = type(juris).__name__

    # if get_session_list is not defined
    if not hasattr(juris, "get_session_list"):
        raise CommandError(f"{scraper}.get_session_list() is not provided")

    scraped_sessions = juris.get_session_list()

    if not scraped_sessions:
        raise CommandError("no sessions from {}.get_session_list()".format(scraper))

    active_sessions = set()
    # copy the list to avoid modifying it
    sessions = set(juris.ignored_scraped_sessions)
    for session in juris.legislative_sessions:
        sessions.add(session.get("_scraped_name", session["identifier"]))
        if session.get("active"):
            active_sessions.add(session.get("identifier"))

    if not active_sessions:
        raise CommandError(f"No active sessions on {scraper}")

    unaccounted_sessions = list(set(scraped_sessions) - sessions)
    if unaccounted_sessions:
        raise CommandError(
            (
                "Session(s) {sessions} were reported by {scraper}.get_session_list() "
                "but were not found in {scraper}.legislative_sessions or "
                "{scraper}.ignored_scraped_sessions."
            ).format(sessions=", ".join(unaccounted_sessions), scraper=scraper)
        )
    stats.send_gauge("active_sessions", len(active_sessions), {"jurisdiction": scraper})
    stats.send_gauge(
        "unaccounted_sessions", len(unaccounted_sessions), {"jurisdiction": scraper}
    )
    stats.send_gauge(
        "ignored_sessions",
        len(juris.ignored_scraped_sessions),
        {"jurisdiction": scraper},
    )
    return active_sessions


def do_update(
    args: argparse.Namespace, other: list[str], juris: State
) -> dict[str, typing.Any]:
    available_scrapers = getattr(juris, "scrapers", {})
    default_scrapers = getattr(juris, "default_scrapers", None)
    scrapers: dict[str, dict[str, str]] = {}

    if not available_scrapers:
        raise CommandError("no scrapers defined on jurisdiction")

    if other:
        # parse arg list in format: (scraper (k:v)+)+
        cur_scraper = None
        for arg in other:
            if "=" in arg:
                if not cur_scraper:
                    raise CommandError("argument {} before scraper name".format(arg))
                k, v = arg.split("=", 1)
                scrapers[cur_scraper][k] = v
            elif arg in juris.scrapers:
                cur_scraper = arg
                scrapers[cur_scraper] = {}
            else:
                raise CommandError(
                    "no such scraper: module={} scraper={}".format(args.module, arg)
                )
    elif default_scrapers is not None:
        scrapers = {s: {} for s in default_scrapers}
    else:
        scrapers = {key: {} for key in available_scrapers.keys()}

    # modify args in-place so we can pass them around
    if not args.actions:
        args.actions = ALL_ACTIONS

    if "import" in args.actions:
        init_django()

    # print the plan
    report = {
        "plan": {"module": args.module, "actions": args.actions, "scrapers": scrapers},
        "start": utils.utcnow(),
    }
    print_report(report)

    if "scrape" in args.actions:
        active_sessions = check_session_list(juris)

    try:
        if "scrape" in args.actions:
            report["scrape"] = do_scrape(juris, args, scrapers, active_sessions)
            stats.send_last_run(
                "last_collection_run_time",
                {
                    "jurisdiction": juris.name,
                    "scrape_type": "scrape",
                },
            )
        # we skip import in realtime mode since this happens via the lambda function
        if "import" in args.actions and not args.realtime:
            report["import"] = do_import(juris, args)
            stats.send_last_run(
                "last_collection_run_time",
                {
                    "jurisdiction": juris.name,
                    "scrape_type": "import",
                },
            )
        report["success"] = True
    except Exception as exc:
        stats.send_counter(
            "scraper_failures_total",
            1,
            {"jurisdiction": juris.name, "scrapers": ",".join(sorted(args.actions))},
        )
        stats.close()
        report["success"] = False
        report["exception"] = exc
        report["traceback"] = traceback.format_exc()
        if "import" in args.actions:
            save_report(report, juris.jurisdiction_id)
        raise
    else:
        finish = utils.utcnow()

        for scrape_type, details in report.get("scrape", {}).items():  # type: ignore
            stats.send_timing(
                "scrape_runtime_secs",
                # datetime - datetime = timedelta object, which has a 'seconds' attribute
                (finish - details["start"]).seconds,
                {
                    "jurisdiction": juris.name,
                    "scrape_type": scrape_type,
                },
            )
            for objtype, num in details["objects"].items():
                stats.send_gauge(
                    "objects_collected",
                    num,
                    {
                        "jurisdiction": juris.name,
                        "scrape_type": scrape_type,
                        "object_type": objtype,
                    },
                )
        for scrape_type, details in report.get("import", {}).items():  # type: ignore
            for import_type in ["insert", "update", "noop"]:
                stats.send_gauge(
                    "objects_imported",
                    details[import_type],
                    {
                        "jurisdiction": juris.name,
                        "scrape_type": scrape_type,
                        "import_type": import_type,
                    },
                )

        if "import" in args.actions:
            save_report(report, juris.jurisdiction_id)

        print_report(report)
        return report


def parse_args() -> tuple[argparse.Namespace, list[str]]:
    parser = argparse.ArgumentParser("openstates", description="openstates CLI")
    parser.add_argument("--debug", action="store_true", help="open debugger on error")
    parser.add_argument(
        "--loglevel",
        default="INFO",
        help=(
            "set log level. options are: "
            "DEBUG|INFO|WARNING|ERROR|CRITICAL "
            "(default is INFO)"
        ),
    )
    # what to scrape
    parser.add_argument("module", type=str, help="path to scraper module")
    for arg in ALL_ACTIONS:
        parser.add_argument(
            "--" + arg,
            dest="actions",
            action="append_const",
            const=arg,
            help="only run {} post-scrape step".format(arg),
        )

    # scraper arguments
    parser.add_argument(
        "--nonstrict",
        action="store_false",
        dest="strict",
        help="skip validation on save",
    )
    parser.add_argument(
        "--fastmode", action="store_true", help="use cache and turn off throttling"
    )

    # settings overrides
    parser.add_argument("--datadir", help="data directory", dest="SCRAPED_DATA_DIR")
    parser.add_argument("--cachedir", help="cache directory", dest="CACHE_DIR")
    parser.add_argument(
        "-r", "--rpm", help="scraper rpm", type=int, dest="SCRAPELIB_RPM"
    )
    parser.add_argument(
        "--timeout", help="scraper timeout", type=int, dest="SCRAPELIB_TIMEOUT"
    )
    parser.add_argument(
        "--no-verify",
        help="skip tls verification",
        action="store_false",
        dest="SCRAPELIB_VERIFY",
    )
    parser.add_argument(
        "--retries", help="scraper retries", type=int, dest="SCRAPELIB_RETRIES"
    )
    parser.add_argument(
        "--retry_wait",
        help="scraper retry wait",
        type=int,
        dest="SCRAPELIB_RETRY_WAIT_SECONDS",
    )

    # realtime mode
    parser.add_argument("--realtime", action="store_true", help="enable realtime mode")

    # process args
    return parser.parse_known_args()


def main() -> int:
    args, other = parse_args()

    # set log level from command line
    handler_level = getattr(logging, args.loglevel.upper(), "INFO")
    settings.LOGGING["handlers"]["default"]["level"] = handler_level  # type: ignore
    logging.config.dictConfig(settings.LOGGING)
    stats.logger.setLevel(handler_level)

    # turn debug on
    if args.debug:
        try:
            debug_module = importlib.import_module("ipdb")
        except ImportError:
            debug_module = importlib.import_module("pdb")

        # turn on PDB-on-error mode
        # stolen from http://stackoverflow.com/questions/1237379/
        # if this causes problems in interactive mode check that page
        def _tb_info(type, value, tb):  # type: ignore
            traceback.print_exception(type, value, tb)
            debug_module.pm()

        sys.excepthook = _tb_info

    logging.info(f"Module: {args.module}")

    juris, module = get_jurisdiction(args.module)

    overrides = {}
    overrides.update(getattr(module, "settings", {}))
    overrides.update(
        {key: value for key, value in vars(args).items() if value is not None}
    )
    with override_settings(settings, overrides):
        report = do_update(args, other, juris)

    stats.close()
    if report.get("success", False):
        return 0
    else:
        return 1


if __name__ == "__main__":
    sys.exit(main())



from ..models import State, Chamber, simple_numbered_districts

TN = State(
    name="Tennessee",
    abbr="TN",
    capital="Nashville",
    capital_tz="America/Chicago",
    fips="47",
    unicameral=False,
    legislature_name="Tennessee General Assembly",
    legislature_organization_id="ocd-organization/2010a137-d178-48fe-a66e-692e97a12fcf",
    executive_name="Office of the Governor",
    executive_organization_id="ocd-organization/81395620-423b-599c-b77c-a0714bff3357",
    division_id="ocd-division/country:us/state:tn",
    jurisdiction_id="ocd-jurisdiction/country:us/state:tn/government",
    url="http://www.capitol.tn.gov/",
    lower=Chamber(
        chamber_type="lower",
        name="House",
        organization_id="ocd-organization/47760e16-8071-49dd-bd9f-354198c8b653",
        num_seats=99,
        title="Representative",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:tn", "lower", 99
        ),
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/d4c442c8-d4d7-4ea9-a3fd-b856b2904d8b",
        num_seats=33,
        title="Senator",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:tn", "upper", 33
        ),
    ),
)



from ..models import State, Chamber, simple_numbered_districts

OH = State(
    name="Ohio",
    abbr="OH",
    capital="Columbus",
    capital_tz="America/New_York",
    fips="39",
    unicameral=False,
    legislature_name="Ohio General Assembly",
    legislature_organization_id="ocd-organization/f100d930-e312-4544-a8d5-710dc5d12ae2",
    executive_name="Office of the Governor",
    executive_organization_id="ocd-organization/04a9a79b-2dc8-4969-b09e-1190dad5b973",
    division_id="ocd-division/country:us/state:oh",
    jurisdiction_id="ocd-jurisdiction/country:us/state:oh/government",
    url="http://www.legislature.state.oh.us/",
    lower=Chamber(
        chamber_type="lower",
        name="House",
        organization_id="ocd-organization/3cb2bb10-f68b-4a7d-a7c5-41986e71ff3a",
        num_seats=99,
        title="Representative",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:oh", "lower", 99
        ),
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/401c065d-0f85-4c93-b87b-1675a7ef6c1a",
        num_seats=33,
        title="Senator",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:oh", "upper", 33
        ),
    ),
)



from ..models import State, Chamber, simple_numbered_districts

ID = State(
    name="Idaho",
    abbr="ID",
    capital="Boise",
    capital_tz="America/Denver",
    fips="16",
    unicameral=False,
    legislature_name="Idaho State Legislature",
    legislature_organization_id="ocd-organization/654d4d38-e413-4307-a420-6f1ee2594673",
    executive_name="Office of the Governor",
    executive_organization_id="ocd-organization/9bbaab73-ae4e-5d25-9f32-c3af64a77829",
    division_id="ocd-division/country:us/state:id",
    jurisdiction_id="ocd-jurisdiction/country:us/state:id/government",
    url="http://www.legislature.idaho.gov",
    lower=Chamber(
        chamber_type="lower",
        name="House",
        organization_id="ocd-organization/7a59916c-201a-49b2-9288-776ffea9f6cb",
        num_seats=70,
        title="Representative",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:id", "lower", 35, num_seats=2
        ),
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/e8c4c8ae-7ba0-4116-9865-5d34fcc6e681",
        num_seats=35,
        title="Senator",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:id", "upper", 35
        ),
    ),
)
